(in-package "PDDL")

; This file builds on top of the types in types.lisp to define indexing
; structures for domains.  The idea is that the domain can be defined
; independently of indexing, but whenever it is updated its indexes are
; rebuilt.

;(DEPENDS-ON AT-RUN-TIME UNPOP/ UNIFY SETINDEX HACKS/ EXPDT)

(defvar uvar-topcoord* (topcoord #'qvar-p))
(declaim (special topcoord*))

(setq topcoord* uvar-topcoord*)

(defvar try-rehash* t)

(defvar break-on-rehash-failure* nil)

(defvar rehash-attempts* 0)
(defvar rehash-failures* 0) 

(defmacro try-rehash (rehasher pat avoid avpat &rest args)
   `(cond (try-rehash* 
	   (setf rehash-attempts* (+ rehash-attempts* 1))
	   (let ((dif (discrim ,pat ,avpat uvar-topcoord*)))
	      (cond ((and dif
			  (,rehasher ,pat ,avoid dif ,@args)))
                    (t
                     (cond (break-on-rehash-failure*
                            (cerror "I will proceed"
                               "Rehash failed after no match of ~s and ~s"
                               ,pat ,avoid)))
		     (setf rehash-failures* (+ rehash-failures* 1))))))))


(defstruct (indexed-domain
	      (:include domain)
	      (:print-function
	         (lambda (id srm k)
		    (declare (ignore k))
		    (format srm "#<Indexed domain ~s>"
			    (domain-name id)))))
   (fill-time 0 :type integer)
   (axioms nil)  ; :type exp_index    get packages figured out
;   (action-defns '()) ; alist of functor, defn pairs
   ;(methods nil) ; :type exp_index  ; actions & methods
   (actions nil) ; :type exp_index  ; uniquified action terms
   (occasions nil) 
)

(defun indexed-domain-with-name (name create)
   (let ((dom (try-domain-with-name name create)))
      (cond (dom
	     (domain-indexify dom))
	    (t nil))))

(defvar idom*)

(defstruct (initialized-problem
              (:include problem)
              (:print-function
	         (lambda (ip srm k)
		    (declare (ignore k))
		    (format srm "#<Initialized problem ~s>"
			    (problem-name ip)))))
   expanded-goal)

(defun domain-indexify (dom)
   (cond ((indexed-domain-p dom)
	  dom)
	 (t
	  (let ((domname (domain-name dom)))
             (multiple-value-bind (dset dget)
                                  (domain-name-procs dom)
	        (let ((value-of-name (funcall dget)))
	           (cond ((indexed-domain-p value-of-name)
		          value-of-name)
		         (t
		          (let ((idom (make-indexed-domain
                                         :fill-time 0
                                         :name (domain-name dom)
                                         :generation (domain-generation dom)
                                         :ancestors (domain-ancestors dom)
                                         :parents (domain-parents dom)
                                         :requirements (domain-requirements dom)
                                         :local-bdgs (domain-local-bdgs dom)
                                         :inherited-bdgs (domain-inherited-bdgs
                                                            dom))))
                             (funcall dset idom)
			     idom)))))))))

(defun fetch-axioms (pat idom)
   ;(indexed-domain-make-current idom)
   (exp-fetch pat (indexed-domain-axioms idom) nil))

;(defun fetch-methods (pat idom)
;   (indexed-domain-make-current idom)
;   (exp-fetch pat (indexed-domain-methods idom) nil))

(defun indexed-domain-make-current (idom)
   (domain-make-current idom)
   (cond ((or (null (indexed-domain-axioms idom))
	      ;(null (indexed-domain-methods idom))
	      (let ((g (indexed-domain-fill-time idom)))
		 (dolist (a (domain-ancestors idom) nil)
		    (cond ((> (domain-generation a) g)
			   (return t))))))
	  ; Some ancestor redefined; rebuild indexes
	  (setf (indexed-domain-axioms idom)
		(exp-index-init #'axiom-consequent))
;	  (setf (indexed-domain-methods idom)
;		(exp-index-init #'termed-action-term))
	  (setf (indexed-domain-generation idom)
		(next-generation))
	  (indexed-domain-fill idom))))

(defun indexed-domain-fill (idom)
   (let ((linkspots '()))
      (dolist (dom (indexed-domain-ancestors idom))
          (dolist (bdg (domain-local-bdgs dom))
             (let ((v (pddl-bdg-val (cadr bdg))))
                (cond ((rule-group-p v)
                       (setq linkspots
                             (nconc (rule-group-index v idom)
                                    linkspots)))
                      ((pddl-type-p v)
                       ;(format t "Indexing ~s in ~s~%" v dom)
                       (type-index v idom))))))
      (dolist (s linkspots)
         (prelink s idom)))
   (setf (indexed-domain-fill-time idom)
	 (next-generation)))

(defun rule-group-index (rg idom)
  ;(format t "Indexing ~s ~%" rg)
  (let ((linkspots '()))
     (dolist (def (rule-group-rules rg))
        (cond ((axiom-p def)
               ;(format t "    Indexing ~s~%" def)
               (multiple-value-bind (expant ls)
                                    (expand-deductive-macros
                                        (axiom-antecedent def)
                                        idom)
                  (setq linkspots (nconc ls linkspots))
                  (exp-ob-index
                     (cond ((procedural-axiom-p def)
                            (make-procedural-axiom
                               :vars (axiom-vars def)
                               :antecedent expant
                               :consequent (axiom-consequent def)
                               :fcn (procedural-axiom-fcn def)))
                           (t
                            (make-axiom
                               :vars (axiom-vars def)
                               :antecedent expant
                               :consequent (axiom-consequent def))))
                     (indexed-domain-axioms idom)
                     t)))
              ((action-defn-p def)
               (let ((fctr (action-defn-functor def)))
                  (let ((fctrdef (find-domain-bdg-val fctr idom)))
                     (cond ((action-functor-p fctrdef)
                            ;(format t "Functor ~s ~s~%" fctr fctrdef)
                            (multiple-value-bind
                                            (defn ls)
                                            (action-defn-expand-deductive-macros
                                                 def idom)
                               (pddl-rebind
                                  fctr
                                  (make-action-functor
                                      :name fctr
                                      :argtypes (functor-argtypes fctrdef)
                                      :defn defn)
                                  idom)
                               (setq linkspots (nconc ls linkspots))))
                           (t
                            (error "Incoherent action defn ~s" def))))))))
     linkspots))

(defun action-defn-expand-deductive-macros (def dom)
  (multiple-value-bind (exp-pre pre-ls)
                       (expand-deductive-macros 
                          (action-defn-precond def)
                          dom)
     ;(nisp::out "Precond = " exp-pre)
     (multiple-value-bind (exp-eff eff-ls)
                          (effect-expand-deductive-macros
                             (action-defn-effect def)
                             dom)
         (values
            (make-action-defn
               :functor (action-defn-functor def)
               :term (action-defn-term def)
               :args (action-defn-args def)
               :vars (action-defn-vars def)
               :precond exp-pre
               :effect exp-eff
               :maintain (action-defn-maintain def)
               :has-methods (action-defn-has-methods def)
               :only-in-expansions (action-defn-only-in-expansions def))
            (nconc pre-ls eff-ls)))))

(defun effect-expand-deductive-macros (effect dom)
     (case (car effect)
        (and
           (let ((r '()) (linkspots '()))
              (dolist (e (cdr effect))
                 (multiple-value-bind (exp-e ls)
                                      (effect-expand-deductive-macros e dom)
                    (push exp-e r)
                    (setq linkspots (nconc ls linkspots))))
              (values `(and ,@(reverse r))
                      linkspots)))
        ((not change)
         (values effect '()))
        (when
         (multiple-value-bind (exp-a ls)
                              (expand-deductive-macros (cadr effect) dom)
            (values `(when ,exp-a ,(caddr effect))
                    ls)))
        (forall
         (multiple-value-bind (exp-e ls)
                              (effect-expand-deductive-macros (caddr effect)
                                                              dom)
            (values `(forall ,(cadr effect) ,exp-e)
                    ls)))
        (t
         (values effect '()))))

; Find domain-specific deductive macros and expand.
; Second value is list of places that need to be "prelinked."
(defun expand-deductive-macros (exp idom)
   (cond ((or (atom exp) (qvar-p exp))
          (values exp '()))
         (t
          (let ((fcn (car exp)))
             (cond ((member fcn '(and or not when change imply) :test #'eq)
                    (multiple-value-bind (a ls)
                                         (list-expand-deductive-macros
                                             (cdr exp) idom)
                       (values `(,fcn ,@a)
                               ls)))
                   ((eq fcn 'exists)
                    ; Skolemize.  Note that we flush type info
                    (expand-deductive-macros (caddr exp) idom))
                   ((eq fcn 'forall)
                        ;(member fcn '(forall exists) :test #'eq)
                    (multiple-value-bind (e ls)
                                         (expand-deductive-macros
                                            (caddr exp) idom)
                       (values `(,fcn ,(cadr exp) ,e)
                               ls)))
                   (t
                    (let ((b (find-domain-bdg fcn idom)))
                       (cond (b
                              (let ((v (pddl-bdg-val b)))
                                 (cond ((predicate-functor-p v)
                                        (let ((fctr (pddl-bdg-val b)))
                                           (cond ((predicate-functor-macro fctr)
                                                  (expand-deductive-macros
                                                      (funcall
                                                         (predicate-functor-macro
                                                            fctr)
                                                         exp)
                                                      idom))
                                                 ((predicate-functor-rules-only
                                                     fctr)
                                                  (let ((pat (copy-list exp)))
                                                     (values pat
                                                             (list pat))))
                                                 (t
                                                  (list-expand-deductive-macros
                                                     exp idom)))))
                                       ((pddl-type-p v)
                                        (values exp '()))
                                       (t
                                        (error "Crazy predicate ~s"
                                               exp)))))
                             (t
                              (error "Unknown predicate ~s" exp))))))))))

(defun list-expand-deductive-macros (l idom)
   (let ((r '()) (linkspots '()))
      (dolist (e l)
         (multiple-value-bind (exp-e ls)
                              (expand-deductive-macros e idom)
            (setq linkspots (nconc ls linkspots))
            (push exp-e r)))
      (values (reverse r) linkspots)))

;type-index creates axioms for sub/super type links
(defun type-index (ty dom)
   (let ((handled '())
	 (typename (pddl-type-defn ty)))
      (labels ((super-if-sub (ty)
		  (cond ((not (member ty handled :test #'eq))
			 (push ty handled)
			 (dolist (par (pddl-type-parents ty))
			    (cond ((not (eq (pddl-type-defn par)
					    'object))
				   (let ((pp (type-predify
						(pddl-type-defn par)
						(make-qvar 'x))))
				      (exp-ob-index
					 (make-procedural-axiom
					    :vars (list '?x)
					    :fcn
					       #'(lambda (qr rid env sit dom)
						         (declare (ignore qr))
						    (collect-subtype-elements
						        ty rid env sit dom))
					    :consequent pp)
					 (indexed-domain-axioms dom)
					 t)
				      (super-if-sub par)))))))))
	  (cond ((and (symbolp typename)
		      (not (eq typename 'object)))
		 (let ((pp (type-predify typename (make-qvar 'x))))
		    (exp-ob-index
		       (make-procedural-axiom
			  :vars (list '?x)
			  :antecedent '(and)
			  :fcn 
			     #'(lambda (qr rid env sit dom)
				       (declare (ignore qr sit))
				  (constant-type-predication
				     ty rid env dom))
			  :consequent pp)
		       (indexed-domain-axioms dom)
		       t))))
	  (super-if-sub ty))))

;(defun action-termify (act-or-meth)
;   (make-termed-action
;      :term (cond ((action-defn-p act-or-meth)
;		   (cons (action-defn-functor act-or-meth)
;			 (var-constraints-args
;			    (action-defn-args act-or-meth))))
;		  (t
;		   (cons (action-method-functor act-or-meth)
;			 (var-constraints-args
;			    (action-method-args act-or-meth)))))
;      :act-or-meth act-or-meth))

(defun prelink (exp idom)
  (let ((pat (copy-list exp))
        (rules (fetch-axioms exp idom)))
     (setf (cdr exp) (list pat))
     (setf (car exp) 'prelinked)
     (setf (cddr exp)
           rules)))

(defun var-constraints-args (cl)
   (labels ((find-first-uvar (x)
	       (cond ((qvar-p x) x)
		     ((atom x) nil)
		     (t
		      (or (find-first-uvar (car x))
			  (find-first-uvar (cdr x)))))))
      (mapcar #'find-first-uvar
	      cl)))

; The adds and deletes fields are currently not used.  They are reserved
; for storing inevitable effects of an action (i.e., those that occur no
; matter what the situation).
(defstruct progress-spec
   action
   (adds '*unknown)
   (deletes '*unknown))

(defun uniquify-action (a context)
   (let ((index (indexed-domain-place-action-index context)))
      (dolist (cand (exp-fetch a index t)
		    (let ((new (make-progress-spec :action a)))
		       (exp-ob-index new index t)
		       (progress-spec-action new)))
	 (cond ((equal (progress-spec-action cand)
		       a)
		(return (progress-spec-action cand)))))))

(defun indexed-domain-place-action-index (idom)
   (let ((ai (indexed-domain-actions idom)))
      (cond ((not ai)
	     (setq ai (exp-index-init #'progress-spec-action))
	     (setf (indexed-domain-actions idom)
		   ai)))
      ai))

(defun occasion-prop (occ) OCC)

(defun uniquify-occasion (occ context)
  (let ((index (indexed-domain-place-occasion-index context)))
     (dolist (cand (exp-fetch occ index t)
		   (progn
		       (exp-ob-index occ index t)
		       occ))
	(cond ((equal cand OCC)
	       (return cand))))))

(defun indexed-domain-place-occasion-index (idom)
   (let ((oi (indexed-domain-occasions idom)))
      (cond ((not oi)
	     (setq oi (exp-index-init #'occasion-prop))
	     (setf (indexed-domain-occasions idom)
		   oi)))
      oi))

(defun uniquify-situation (sit path init)
   (let ((ind (initial-situation-sit-index init))
	 (conts (exp_index-contents sit nil)))
      (cond ((not ind)
	     (setq ind (make-eq_setindex #'get-situation-contents))
	     (eq-index-add init ind)
	     (setf (initial-situation-sit-index init)
		   ind)))
      (dolist (cs (elts-eq-index-fetch conts ind)
		  (let ((new (make-subsequent-situation
				:index sit
				:contents conts
				:path path
				:init init)))
		    (eq-index-add new ind)
		    new))
	(cond ((and (exp-index-subconts sit (situation-index cs))
		    (exp-index-subconts (situation-index cs) sit))
	       (return cs))))))

(defstruct (sitsuccspec
	      (:type list))
   act
   sit
   (opt nil :type boolean)
   (adds '() :type list)
   (deletes '() :type list))

(defun initial-situation-domain-indexify (isit)
   (let ((idom (initial-situation-domain isit)))
     (cond ((not (indexed-domain-p idom))
	    (setf idom (domain-indexify idom))
	    (setf (initial-situation-domain isit)
		  idom)))
     (indexed-domain-make-current idom)
     idom))


; Initialize timeless index of problem with all the timeless facts
; from domain and parent sit.
(defun problem-initialize (prob)
   (situation-initialize (problem-sit prob))
   (cond ((not (initialized-problem-p prob))
          (setq prob (make-initialized-problem
                        :name (problem-name prob)
                        :sit (problem-sit prob)
                        :goal (problem-goal prob)
                        :expansion (problem-expansion prob)
                        :length (problem-length prob)))
          (set-global-pddl-symbol (problem-name prob) prob)))
   (setf (initialized-problem-expanded-goal prob)
         (expand-deductive-macros (initialized-problem-goal prob)
                                  (problem-domain prob)))
   prob)
               
(defun situation-initialize (sit)
      (let ((idom (initial-situation-domain-indexify sit)))
	 (indexed-domain-make-current idom)
	 (cond ((or (not (initial-situation-timeless-index sit))
		    (not (situation-index sit))
		    (some #'(lambda (a)
			   (> (domain-generation a)
			      (initial-situation-generation sit)))
		      (domain-ancestors idom)))
		(let ((timeless-ind (exp-index-init #'occasion-prop))
		      (timeful-ind (exp-index-init #'occasion-prop))
		      (checked-domains '()))
		   (setf (initial-situation-timeless-index sit)
			 timeless-ind)
		   (setf (initial-situation-index sit)
			 timeful-ind)
		   (setf (initial-situation-sit-index sit)
			 nil)
                   (setf (situation-succs sit) '())
		   (labels ((check-next (dom)
			       (cond ((not (member dom checked-domains
						   :test #'eq))
				      (push dom checked-domains)
				      ;(format t "Checking ~s ~%" dom)
				      (dolist (sup (domain-ancestors dom))
					 (check-next sup))
				      (grab-from-domain dom))))
                            (constant-to-timeless (name ty)
                               (let ((p (type-predify
                                           (pddl-type-defn ty)
                                           name)))
                                  (cond ((not (member (car p)
                                                      '(object or)
                                                      :test #'eq))
                                         (exp-ob-index
                                            (uniquify-occasion p idom)
                                            timeless-ind
                                            t)))))
			    (grab-from-domain (pdom)
			       ;(format t "~% Grabbing from ~s :" pdom)
			       (dolist (p (domain-timeless pdom))
				  (add-or-subtract p t))
			       (dolist (v (domain-local-bdgs pdom))
				  (let ((b (pddl-bdg-val (cadr v))))
				     (cond ((constant-p b)
                                            (constant-to-timeless
                                                (car v)
                                                (constant-type b)))
                                           ((domain-var-p b)
                                            (constant-to-timeless
                                                (car v)
                                                (domain-var-type b)))))))
			    (add-or-subtract (p whether-add)
			       (cond ((eq (car p) 'not)
				      (add-or-subtract
				         (cadr p)
					 (not whether-add)))
				     (t
				      (exp-ob-index
				          (uniquify-occasion p idom)
					  (cond ((type-name-p (car p) idom)
						 timeless-ind)
						(t timeful-ind))
					  whether-add)))))
		      (check-next idom)
		      (do ((next-sit sit (initial-situation-parent next-sit)))
			  ((not next-sit))
			 (dolist (c (initial-situation-delta next-sit))
			    (add-or-subtract c t)))
		      (setf (initial-situation-generation sit)
			    (next-generation))))))
         sit))

(defun get-situation-contents (sit)
   (let ((c (situation-contents sit)))
      (cond ((eq c '*uncomputed)
	     (setq c (exp_index-contents
		         (get-situation-index sit)
			 nil))
	     (setf (situation-contents sit) c)))
      c))
      
; Build it if necessary 
(defun get-situation-index (sit)
      (or (situation-index sit)
	  (cond ((initial-situation-p sit)
		 (situation-index (situation-initialize sit)))
		(t
		 (error "Non-initial situation with no index ~s" sit)))))

;	  (let ((ind (exp-index-init #'occasion-prop)))
;	     (dolist (occ (situation-contents sit))
;		(exp-ob-index occ ind t))
;	     (setf (situation-index sit) ind)
;	     ind)))



(setf universal-ancestors*
      (mapcar #'domain-indexify universal-ancestors*))

(defun all-rule-groups (dom)
   (let ((groups '())
	 (checked-domains '()))
      (labels ((check-next (dom)
		  (cond ((not (member dom checked-domains :test #'eq))
			 (push dom checked-domains)
			 (dolist (bdg (domain-local-bdgs dom))
			    (let ((x (pddl-bdg-val (cadr bdg))))
			      (cond ((rule-group-p x)
				     (push x groups)))))
			 (dolist (sup (domain-ancestors dom))
			    (check-next sup))))))
	  (check-next dom)
	  groups)))

(defun action-methods (fctr idom)
   (let ((gl (all-rule-groups idom))
	 (meths '()))
      (dolist (g gl)
	 (dolist (x (rule-group-rules g))
	    (cond ((and (action-method-p x)
			(eq (action-method-functor x)
			    fctr))
		   (push x meths)))))
      meths))

(defmacro def-deductive-macro (name domain args &rest body)
   `(define-ded-macro
       ',name
       (must-domain-with-name ',domain)
       #'(lambda ,args ,@body)))
  
(defun must-domain-with-name (name)
   (let ((dom (try-domain-with-name name nil)))
      (cond ((not dom)
             (error "Undefined domain ~s" name)))
      dom))
                    
(defun define-ded-macro (name dom fcn)
   (let ((fctrb (find-domain-bdg name dom)))
     (cond ((and fctrb (predicate-functor-p (pddl-bdg-val fctrb)))
            (setf (domain-generation dom)
                  (next-generation))
            (setf (predicate-functor-macro (pddl-bdg-val fctrb))
                  fcn))
           (t
            (error "Can't define ~s as deductive macro (must be declared as predicate)"
                   name)))))
